/**
 * Copyright (C) 2008 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h> /* R_OK */

#include <glib/gstdio.h>
#include <gdk/gdkx.h>

#include <vlc/vlc.h>

#include "ltop-video.h"
#include "debug.h"

struct _LtopVideoPrivate {
    libvlc_instance_t     *vlc_inst;
    libvlc_media_player_t *vlc_mp;
    libvlc_exception_t     vlc_ex;

    GThread *thread;

    gint running;
};

static void ltop_video_media_init (LtopMediaInterface *iface);
static void ltop_video_class_init (LtopVideoClass *class);
static void ltop_video_init       (LtopVideo *video);
static void ltop_video_finalize   (GObject *object);

static void ltop_video_run        (LtopMedia *media);
static void ltop_video_attach_to_window (LtopMedia *media,
                                         GdkWindow *window);

G_DEFINE_TYPE_EXTENDED (LtopVideo,
                        ltop_video,
                        G_TYPE_OBJECT,
                        0,
                        G_IMPLEMENT_INTERFACE (LTOP_TYPE_MEDIA,
                                               ltop_video_media_init));

/* interface init */
static void
ltop_video_media_init (LtopMediaInterface *iface)
{
    iface->run = ltop_video_run;
    iface->attach_to_window = ltop_video_attach_to_window;
}

static void
ltop_video_class_init (LtopVideoClass *class)
{
    g_type_class_add_private (class, sizeof (LtopVideoPrivate));

    /* init thread system if not already done */
    if (!g_thread_supported ())
        g_thread_init (NULL);

    GObjectClass *object_class = (GObjectClass *) class;

    object_class->finalize = ltop_video_finalize;
}

static void
ltop_video_init (LtopVideo *video)
{
    video->priv = G_TYPE_INSTANCE_GET_PRIVATE (video,
                                               LTOP_TYPE_VIDEO,
                                               LtopVideoPrivate);

    /* vlc init arguments */
    const gchar * const vlc_args[] = {
        "-I", "dummy",
        "--ignore-config",
        "--no-video-title-show",
    };

    libvlc_exception_init (&video->priv->vlc_ex);
    video->priv->vlc_inst = libvlc_new (
            sizeof(vlc_args) / sizeof(vlc_args[0]),
            vlc_args, &video->priv->vlc_ex);
    libvlc_exception_clear (&video->priv->vlc_ex);
    g_atomic_int_set (&video->priv->running, 0);
}

static void
ltop_video_finalize (GObject *object)
{
    DBG("entering");
    LtopVideo *video = LTOP_VIDEO (object);

    libvlc_media_player_release (video->priv->vlc_mp);
    libvlc_release (video->priv->vlc_inst);

    (*G_OBJECT_CLASS (ltop_video_parent_class)->finalize) (object);
    DBG("leaving");
}

static void
media_end_reached (const libvlc_event_t *e, void *data)
{
    LtopVideo *video = LTOP_VIDEO (data);
    g_atomic_int_set (&video->priv->running, 0);
}

static gpointer
run_video_thread (gpointer data)
{
    LtopVideo *video = LTOP_VIDEO (data);
    libvlc_media_player_play (video->priv->vlc_mp, &video->priv->vlc_ex);
    libvlc_exception_clear (&video->priv->vlc_ex);

    while (1)
    {
        sleep (0.5);
        if (!g_atomic_int_get (&video->priv->running)) break;
    }
    libvlc_media_player_stop (video->priv->vlc_mp, &video->priv->vlc_ex);
    libvlc_exception_clear (&video->priv->vlc_ex);

    DBG("leaving thread");
    g_thread_exit ((gpointer) TRUE);

    return (gpointer) TRUE;
}

GObject *
ltop_video_new (const gchar *filename)
{
    g_return_val_if_fail (filename != NULL, NULL);
    g_return_val_if_fail (!g_access (filename, R_OK), NULL);

    LtopVideo *video = g_object_new (LTOP_TYPE_VIDEO, NULL);

    /* load the video file */
    libvlc_media_t *m = libvlc_media_new (video->priv->vlc_inst,
                                          filename,
                                          &video->priv->vlc_ex);
    libvlc_exception_clear (&video->priv->vlc_ex);
    video->priv->vlc_mp = libvlc_media_player_new_from_media (m,
                                          &video->priv->vlc_ex);
    libvlc_exception_clear (&video->priv->vlc_ex);

    /* disable audio (TODO: make this optional) */
    libvlc_audio_set_mute (video->priv->vlc_inst, 1,
                           &video->priv->vlc_ex);
    libvlc_exception_clear (&video->priv->vlc_ex);

    /* handle events */
    libvlc_event_manager_t *em = libvlc_media_player_event_manager (
                                           video->priv->vlc_mp,
                                           &video->priv->vlc_ex);
    libvlc_exception_clear (&video->priv->vlc_ex);
    libvlc_event_attach (em,
                         libvlc_MediaPlayerEndReached,
                         (libvlc_callback_t) media_end_reached,
                         (void *) video,
                         &video->priv->vlc_ex);
    libvlc_exception_clear (&video->priv->vlc_ex);

    return G_OBJECT (video);
}

static void
ltop_video_run (LtopMedia *media)
{
    g_return_if_fail (LTOP_IS_MEDIA (media));

    LtopVideo *video = LTOP_VIDEO (media);

    if (g_atomic_int_get (&video->priv->running))
        return;

    /* no thread started, no atomic operation needed */
    video->priv->running = 1;
    DBG("starting video thread");
    video->priv->thread = g_thread_create (run_video_thread,
                                           (gpointer) video,
                                           TRUE,
                                           NULL);
    g_thread_join (video->priv->thread);
    DBG("thread joined\n");
    ltop_media_job_done (LTOP_MEDIA (video));
}

static void
ltop_video_attach_to_window (LtopMedia *media,
                             GdkWindow *window)
{
    g_return_if_fail (LTOP_IS_MEDIA (media));
    g_return_if_fail (GDK_IS_WINDOW (window));

    LtopVideo *video = LTOP_VIDEO (media);

    DBG("attaching video to gdkwindow");
    libvlc_drawable_t drawable = GDK_DRAWABLE_XID (GDK_DRAWABLE (window));
    libvlc_media_player_set_drawable (video->priv->vlc_mp,
                                      drawable,
                                      &video->priv->vlc_ex);
    libvlc_exception_clear (&video->priv->vlc_ex);
}

